@@ -1,15 +1,15 @@
 # -*- Mode: cperl; cperl-indent-level: 4 -*-
-# $Id: Straps.pm,v 1.1.2.10 2001/12/18 03:20:10 schwern Exp $
+# $Id: Straps.pm,v 1.16 2003/02/02 05:27:44 schwern Exp $
 
 package Test::Harness::Straps;
 
 use strict;
 use vars qw($VERSION);
 use Config;
-$VERSION = '0.05';
+$VERSION = '0.14';
 
 use Test::Harness::Assert;
-
+use Test::Harness::Iterator;
 
 # Flags used as return values from our methods.  Just for internal 
 # clarification.
@@ -30,7 +30,7 @@ Test::Harness::Straps - detailed analysis of test results
   my $strap = Test::Harness::Straps->new;
 
   # Various ways to interpret a test
-  my %results = $strap->analyze($name, @test_output);
+  my %results = $strap->analyze($name, \@test_output);
   my %results = $strap->analyze_fh($name, $test_filehandle);
   my %results = $strap->analyze_file($test_file);
 
@@ -44,6 +44,9 @@ Test::Harness::Straps - detailed analysis of test results
 
 =head1 DESCRIPTION
 
+B<THIS IS ALPHA SOFTWARE> in that the interface is subject to change
+in incompatible ways.  It is otherwise stable.
+
 Test::Harness is limited to printing out its results.  This makes
 analysis of the test results difficult for anything but a human.  To
 make it easier for programs to work with test results, we provide
@@ -51,6 +54,10 @@ Test::Harness::Straps.  Instead of printing the results, straps
 provide them as raw data.  You can also configure how the tests are to
 be run.
 
+The interface is currently incomplete.  I<Please> contact the author
+if you'd like a feature added or something change or just have
+comments.
+
 =head2 Construction
 
 =over 4
@@ -86,7 +93,8 @@ Initialize the internal state of a strap to make it ready for parsing.
 sub _init {
     my($self) = shift;
 
-    $self->{_is_vms} = $^O eq 'VMS';
+    $self->{_is_vms}   = $^O eq 'VMS';
+    $self->{_is_win32} = $^O eq 'Win32';
 }
 
 =end _private
@@ -112,6 +120,14 @@ L<Results>.
 sub analyze {
     my($self, $name, $test_output) = @_;
 
+    my $it = Test::Harness::Iterator->new($test_output);
+    return $self->_analyze_iterator($name, $it);
+}
+
+
+sub _analyze_iterator {
+    my($self, $name, $it) = @_;
+
     $self->_reset_file_state;
     $self->{file} = $name;
     my %totals  = (
@@ -122,71 +138,99 @@ sub analyze {
                    todo     => 0,
                    skip     => 0,
                    bonus    => 0,
-                   
+
                    details  => []
                   );
 
-
-    foreach my $line (@$test_output) {
-        my %result = ();
-        
-        $self->{line}++;
-        
-
-        if( $self->_is_header($line) ) {
-            $self->{saw_header}++;
-
-            $totals{max} += $self->{max};
-
-#            $self->_callback('saw_header', $line);
-        }
-        elsif( $self->_is_test($line, \%result) ) {
-            $totals{seen}++;
-            $result{number} = $self->{'next'} unless $result{number};
-
-            my $pass = $result{ok};
-
-            # special case for VMS, sometimes the 'not ' and the 'ok' are
-            # on different lines if you do this:
-            #   print "not " unless $test;
-            #   print "ok $num\n";
-            if( $self->{_is_vms}                             &&
-                $self->{_saw_lone_not} == $self->{line} - 1  &&
-                $result{ok}
-              ) 
-            {   
-                $result{ok} = 0;
-            }
-
-            $result{type} = 'todo' if $self->{todo}{$result{number}};
-
-            if( $result{type} eq 'todo' ) {
-                $totals{todo}++;
-                $pass = 1;
-                $totals{bonus}++ if $result{ok}
-            }
-            elsif( $result{type} eq 'skip' ) {
-                $totals{skip}++;
-                $pass = 1;
-            }
-
-            $totals{ok}++ if $pass;
-
-            $totals{details}[$result{number} - 1] = 
-              {$self->_detailize($pass, \%result)};
-
-            $self->{'next'}++;
-
-            # XXX handle counter mismatch
-
-        }
+    # Set them up here so callbacks can have them.
+    $self->{totals}{$name}         = \%totals;
+    while( defined(my $line = $it->next) ) {
+        $self->_analyze_line($line, \%totals);
+        last if $self->{saw_bailout};
     }
 
-    $self->{totals}{$name} = \%totals;
+    $totals{skip_all} = $self->{skip_all} if defined $self->{skip_all};
+
+    my $passed = ($totals{max} == 0 && defined $totals{skip_all}) ||
+                 ($totals{max} && $totals{seen} &&
+                  $totals{max} == $totals{seen} && 
+                  $totals{max} == $totals{ok});
+    $totals{passing} = $passed ? 1 : 0;
+
     return %totals;
 }
 
 
+sub _analyze_line {
+    my($self, $line, $totals) = @_;
+
+    my %result = ();
+
+    $self->{line}++;
+
+    my $type;
+    if( $self->_is_header($line) ) {
+        $type = 'header';
+
+        $self->{saw_header}++;
+
+        $totals->{max} += $self->{max};
+    }
+    elsif( $self->_is_test($line, \%result) ) {
+        $type = 'test';
+
+        $totals->{seen}++;
+        $result{number} = $self->{'next'} unless $result{number};
+
+        # sometimes the 'not ' and the 'ok' are on different lines,
+        # happens often on VMS if you do:
+        #   print "not " unless $test;
+        #   print "ok $num\n";
+        if( $self->{saw_lone_not} && 
+            ($self->{lone_not_line} == $self->{line} - 1) ) 
+        {
+            $result{ok} = 0;
+        }
+
+        my $pass = $result{ok};
+        $result{type} = 'todo' if $self->{todo}{$result{number}};
+
+        if( $result{type} eq 'todo' ) {
+            $totals->{todo}++;
+            $pass = 1;
+            $totals->{bonus}++ if $result{ok}
+        }
+        elsif( $result{type} eq 'skip' ) {
+            $totals->{skip}++;
+            $pass = 1;
+        }
+
+        $totals->{ok}++ if $pass;
+
+        if( $result{number} > 100000 && $result{number} > $self->{max} ) {
+            warn "Enormous test number seen [test $result{number}]\n";
+            warn "Can't detailize, too big.\n";
+        }
+        else {
+            $totals->{details}[$result{number} - 1] = 
+                               {$self->_detailize($pass, \%result)};
+        }
+
+        # XXX handle counter mismatch
+    }
+    elsif ( $self->_is_bail_out($line, \$self->{bailout_reason}) ) {
+        $type = 'bailout';
+        $self->{saw_bailout} = 1;
+    }
+    else {
+        $type = 'other';
+    }
+
+    $self->{callback}->($self, $line, $type, $totals) if $self->{callback};
+
+    $self->{'next'} = $result{number} + 1 if $type eq 'test';
+}
+
 =item B<analyze_fh>
 
   my %results = $strap->analyze_fh($name, $test_filehandle);
@@ -198,47 +242,72 @@ Like C<analyze>, but it reads from the given filehandle.
 sub analyze_fh {
     my($self, $name, $fh) = @_;
 
-    $self->analyze($name, [<$fh>]);
+    my $it = Test::Harness::Iterator->new($fh);
+    $self->_analyze_iterator($name, $it);
 }
 
 =item B<analyze_file>
 
   my %results = $strap->analyze_file($test_file);
 
-Like C<analyze>, but it reads from the given $test_file.  It will also
-use that name for the total report.
+Like C<analyze>, but it runs the given $test_file and parses it's
+results.  It will also use that name for the total report.
 
 =cut
 
 sub analyze_file {
     my($self, $file) = @_;
 
-    local $ENV{IFS} = '';
+    unless( -e $file ) {
+        $self->{error} = "$file does not exist";
+        return;
+    }
 
-    # XXX set switches
+    unless( -r $file ) {
+        $self->{error} = "$file is not readable";
+        return;
+    }
 
     local $ENV{PERL5LIB} = $self->_INC2PERL5LIB;
 
-    # Is this necessary anymore?
-    my $cmd = $self->{_is_vms} ? "MCR $^X" : $^X;
+    my $cmd = $self->{_is_vms}   ? "MCR $^X" :
+              $self->{_is_win32} ? Win32::GetShortPathName($^X)
+                                 : $^X;
 
     my $switches = $self->_switches($file);
 
     # *sigh* this breaks under taint, but open -| is unportable.
     unless( open(FILE, "$cmd $switches $file|") ) {
-        # XXX fix up this error handling.
         print "can't run $file. $!\n";
         return;
     }
 
-    # XXX simple, stupid, easy for now.
     my %results = $self->analyze_fh($file, \*FILE);
+    my $exit = close FILE;
+    $results{'wait'} = $?;
+    if( $? && $self->{_is_vms} ) {
+        eval q{use vmsish "status"; $results{'exit'} = $?};
+    }
+    else {
+        $results{'exit'} = _wait2exit($?);
+    }
+    $results{passing} = 0 unless $? == 0;
 
     $self->_restore_PERL5LIB();
 
     return %results;
 }
 
+
+eval { require POSIX; &POSIX::WEXITSTATUS(0) };
+if( $@ ) {
+    *_wait2exit = sub { $_[0] >> 8 };
+}
+else {
+    *_wait2exit = sub { POSIX::WEXITSTATUS($_[0]) }
+}
+
+
 =begin _private
 
 =item B<_switches>
@@ -258,8 +327,16 @@ sub _switches {
     my $s = '';
     $s .= " $ENV{'HARNESS_PERL_SWITCHES'}"
       if exists $ENV{'HARNESS_PERL_SWITCHES'};
-    $s .= join " ", qq[ "-$1"], map {qq["-I$_"]} $self->_filtered_INC
-      if $first =~ /^#!.*\bperl.*-\w*([Tt]+)/;
+
+    if ($first =~ /^#!.*\bperl.*\s-\w*([Tt]+)/) {
+        # When taint mode is on, PERL5LIB is ignored.  So we need to put
+        # all that on the command line as -Is.
+        $s .= join " ", qq[ "-$1"], map {qq["-I$_"]} $self->_filtered_INC;
+    }
+    elsif ($^O eq 'MacOS') {
+        # MacPerl's putenv is broken, so it will not see PERL5LIB.
+        $s .= join " ", map {qq["-I$_"]} $self->_filtered_INC;
+    }
 
     close(TEST) or print "can't close $file. $!\n";
 
@@ -282,7 +359,7 @@ sub _INC2PERL5LIB {
     $self->{_old5lib} = $ENV{PERL5LIB};
 
     return join $Config{path_sep}, $self->_filtered_INC;
-}    
+}
 
 =item B<_filtered_INC>
 
@@ -326,7 +403,7 @@ sub _restore_PERL5LIB {
         $ENV{PERL5LIB} = $self->{_old5lib};
     }
 }
-    
+
 
 =end _private
 
@@ -377,7 +454,7 @@ $strap->{skip_all} contains the reason.
 my $Extra_Header_Re = <<'REGEX';
                        ^
                         (?: \s+ todo \s+ ([\d \t]+) )?      # optional todo set
-                        (?: \s* \# \s* ([\w:]+) (.*) )?     # optional skip with optional reason
+                        (?: \s* \# \s* ([\w:]+\s?) (.*) )?     # optional skip with optional reason
 REGEX
 
 sub _is_header {
@@ -392,7 +469,11 @@ sub _is_header {
 
             $self->{todo} = { map { $_ => 1 } split /\s+/, $todo } if $todo;
 
-            $self->{skip_all} = $reason if defined $skip and $skip =~ /^Skip/i;
+            if( $self->{max} == 0 ) {
+                $reason = '' unless defined $skip and $skip =~ /^Skip/i;
+            }
+
+            $self->{skip_all} = $reason;
         }
 
         return $YES;
@@ -416,7 +497,8 @@ result back in %test which will contain:
   type          'todo' or 'skip' (if any)
   reason        why is it todo or skip? (if any)
 
-If will also catch lone 'not' lines and note them in $strap->{_saw_lone_not}.
+If will also catch lone 'not' lines, note it saw them 
+$strap->{saw_lone_not} and the line in $strap->{lone_not_line}.
 
 =cut
 
@@ -461,8 +543,9 @@ sub _is_test {
     else{
         # Sometimes the "not " and "ok" will be on seperate lines on VMS.
         # We catch this and remember we saw it.
-        if( $self->{_is_vms} && $line =~ /^not\s+$/ ) {
-            $self->{_saw_lone_not} = $self->{line};
+        if( $line =~ /^not\s+$/ ) {
+            $self->{saw_lone_not} = 1;
+            $self->{lone_not_line} = $self->{line};
         }
 
         return $NO;
@@ -505,6 +588,10 @@ sub _reset_file_state {
     delete @{$self}{qw(max skip_all todo)};
     $self->{line}       = 0;
     $self->{saw_header} = 0;
+    $self->{saw_bailout}= 0;
+    $self->{saw_lone_not} = 0;
+    $self->{lone_not_line} = 0;
+    $self->{bailout_reason} = '';
     $self->{'next'}       = 1;
 }
 
@@ -517,6 +604,12 @@ sub _reset_file_state {
 
 The %results returned from analyze() contain the following information:
 
+  passing           true if the whole test is considered a pass 
+                    (or skipped), false if its a failure
+
+  exit              the exit code of the test run, if from a file
+  wait              the wait code of the test run, if from a file
+
   max               total tests which should have been run
   seen              total tests actually seen
   skip_all          if the whole test was skipped, this will 
@@ -574,8 +667,11 @@ sub _detailize {
     assert( !(grep !defined $details{$_}, keys %details),
             'test contains the ok and actual_ok info' );
 
+    # We don't want these to be undef because they are often
+    # checked and don't want the checker to have to deal with
+    # uninitialized vars.
     foreach my $piece (qw(name type reason)) {
-        $details{$piece} = $test->{$piece} if $test->{$piece};
+        $details{$piece} = defined $test->{$piece} ? $test->{$piece} : '';
     }
 
     return %details;
@@ -585,6 +681,10 @@ sub _detailize {
 
 =end _private
 
+=head1 EXAMPLES
+
+See F<examples/mini_harness.plx> for an example of use.
+
 =head1 AUTHOR
 
 Michael G Schwern E<lt>schwern@pobox.comE<gt>
